import java.util.Random;

/**
 * Konto.java - Demonstration von Monitoren in Java.
 *
 * @version 980121
 * @author  <A HREF="http://www.inf.fu-berlin.de/~dahm">M. Dahm</A>
 */
public class Konto {
  private int guthaben; // Aktueller Kontostand

  public Konto(int betrag) {
    guthaben = betrag;
  }

  public synchronized void einzahlen(int betrag) {
    guthaben += betrag; // Einzahlung vornehmen
    this.notifyAll();   /* Alle Threads aufwecken, die sich in der Monitor-
			 * Warteschlange _dieses_ Objekts befinden (this).
			 */
    print();            // Kontostand ausgeben
  }
  
  public synchronized void abheben(int betrag) {
    while(guthaben < betrag) { // Solange Thread noch nicht abheben kann
      try {                    // try und catch sind ntig, aber ignorierbar
	print("Abheber wartet");
	this.wait();           // Warte (auf notify)
      } catch (InterruptedException e) {
	System.out.println(e); System.exit(-1);
      }
    }      

    guthaben -= betrag; // Abhebung vornehmen
    print();            // Kontostand ausgeben
  }

  public synchronized void print(String str) {
    System.out.println(str);
  }

  public synchronized void print() {
    print("Guthaben = " + guthaben);
  }

  public static void main(String[] argv) {
    Konto konto = new Konto(100); // Konto-Objekt erzeugen

    // Ein paar Einzahler und Abheber erzeugen
    new Thread((Runnable)new EinAusZahler(1, konto), "Zahler(1)").start();
    new Thread((Runnable)new EinAusZahler(2, konto), "Zahler(2)").start();
    new Thread((Runnable)new EinAusZahler(3, konto), "Zahler(3)").start();
    new Thread((Runnable)new EinAusZahler(4, konto), "Zahler(4)").start();

    // Die Main-Routine tut eigentlich nichts vernnftiges
    while(true) {
      try {
	Thread.sleep(10000);
      } catch (InterruptedException e) {} 
    }
  }
}

class EinAusZahler implements Runnable {
  public static Random rand = new Random(); /* Hilfsobjekt fr
					     * Zufallszahlen. */
  private int   pid;   // Prozess-Nummer
  private Konto konto; // Konto-Objekt fr Ein/Auszahlungen

  EinAusZahler(int pid, Konto konto) {
    this.pid   = pid;
    this.konto = konto;
  }

  public void run() {
    while(true) {
      // Erst ein zufllig lange schlafen
      try {
	Thread.sleep(Math.abs(rand.nextInt()) % 1000);
      } catch (InterruptedException e) {
	System.out.println(e); System.exit(-1);
      }

      int betrag = Math.abs(rand.nextInt());

      if((pid % 2) == 0) { // Einzahler oder Abheber ?
	betrag = betrag % 100 + 1;
 	konto.print("Einzahler(" + pid + ") zahlt " + betrag + " ein.");
	konto.einzahlen(betrag);
      }
      else {
	betrag = betrag % 200 + 1; // Abheber hebt immer etwas mehr ab
	konto.print("Abheber(" + pid + ") hebt " + betrag + " ab.");
	konto.abheben(betrag);
	konto.print("Abheber(" + pid + ") ist fertig mit abheben.");
      }
    }
  }
}

